import { CollabFail, CollabClientRequestType, ManagerCommunication, CollabManagerBroadCastType } from '@bangle.dev/collab-comms';
export { CollabMessageBus } from '@bangle.dev/collab-comms';
import { Step } from '@bangle.dev/pm';
import { isTestEnv, Either } from '@bangle.dev/utils';

const MAX_STEP_HISTORY = 1000;
const LOG$1 = true;
let log$1 = (isTestEnv ? false : LOG$1)
    ? console.debug.bind(console, 'collab-manager/state:')
    : () => { };
class CollabServerState {
    constructor(doc, steps = [], version = 0) {
        this.doc = doc;
        this.steps = steps;
        this.version = version;
        // Trim steps if needed
        if (steps.length > MAX_STEP_HISTORY) {
            this.steps = steps.slice(steps.length - MAX_STEP_HISTORY);
        }
    }
    static addEvents(state, version, steps, clientID) {
        if (CollabServerState.isInvalidVersion(state, version)) {
            return Either.left(CollabFail.InvalidVersion);
        }
        const biggerSteps = steps.map((s) => Object.assign(s, { clientID }));
        if (state.version !== version) {
            return Either.left(CollabFail.OutdatedVersion);
        }
        let newDoc = state.doc;
        const maps = [];
        for (const step of biggerSteps) {
            let result;
            try {
                result = step.apply(newDoc);
            }
            catch (error) {
                console.error(error);
                return Either.left(CollabFail.ApplyFailed);
            }
            if (result.doc == null) {
                return Either.left(CollabFail.ApplyFailed);
            }
            newDoc = result.doc;
            maps.push(step.getMap());
        }
        const newVersion = state.version + biggerSteps.length;
        const newSteps = state.steps.concat(biggerSteps);
        log$1(`${clientID}: addEvents version=${newVersion}`);
        return Either.right(new CollabServerState(newDoc, newSteps, newVersion));
    }
    static getEvents(collabState, version) {
        if (CollabServerState.isInvalidVersion(collabState, version)) {
            return Either.left(CollabFail.InvalidVersion);
        }
        let startIndex = collabState.steps.length - (collabState.version - version);
        if (startIndex < 0) {
            return Either.left(CollabFail.HistoryNotAvailable);
        }
        return Either.right({
            version: collabState.version,
            steps: collabState.steps.slice(startIndex),
        });
    }
    static isInvalidVersion(collabState, version) {
        return version < 0 || version > collabState.version;
    }
}

// This is the default value if one it not specified by the consumer.
const DEFAULT_MANAGER_ID = '@bangle.dev/collab-manager/MANAGER';

const LOG = true;
let log = (isTestEnv ? false : LOG)
    ? console.debug.bind(console, 'collab-manager:')
    : () => { };
class CollabManager {
    constructor(_options) {
        this._options = _options;
        this._abortController = new AbortController();
        this._handleRequest = async (request, id) => {
            // TODO Add validation of request
            // if (!request.body.docName) {
            //   throw new Error('docName is required');
            // }
            log(`id=${id} userId=${request.body.userId} received request=${request.type}, `);
            switch (request.type) {
                case CollabClientRequestType.GetDocument: {
                    return this._handleGetDocument(request, id);
                }
                case CollabClientRequestType.PullEvents: {
                    return this._handlePullEvents(request, id);
                }
                case CollabClientRequestType.PushEvents: {
                    return this._handlePushEvents(request, id);
                }
                default: {
                    throw new Error('Unknown request type');
                }
            }
        };
        this._instances = new Map();
        this.managerId = _options.managerId || DEFAULT_MANAGER_ID;
        if (_options.collabMessageBus) {
            this._serverCom = new ManagerCommunication(this.managerId, _options.collabMessageBus, this._handleRequest, this._abortController.signal);
        }
    }
    destroy() {
        this._abortController.abort();
    }
    getAllDocNames() {
        return new Set(this._instances.keys());
    }
    getCollabState(docName) {
        var _a;
        return (_a = this._instances.get(docName)) === null || _a === void 0 ? void 0 : _a.collabState;
    }
    isDestroyed() {
        return this._abortController.signal.aborted;
    }
    // removes collab state entry associated with docName
    removeCollabState(docName) {
        this._instances.delete(docName);
    }
    async _createInstance(docName) {
        const initialCollabState = await this._options.getInitialState(docName);
        // this takes care of edge case where another instance is created
        // while we wait on `getInitialState`.
        let instance = this._instances.get(docName);
        if (instance) {
            return Either.right(instance);
        }
        if (!initialCollabState) {
            return Either.left(CollabFail.DocumentNotFound);
        }
        instance = new Instance(docName, this._options.schema, initialCollabState, this._options.applyState);
        this._instances.set(docName, instance);
        return Either.right(instance);
    }
    /**
     * Get an instance of a document or creates a new one
     * if none exists.
     * @param docName
     * @returns
     */
    async _getInstance(docName) {
        const instance = this._instances.get(docName);
        if (instance) {
            instance.lastActive = Date.now();
            return Either.right(instance);
        }
        return this._createInstance(docName);
    }
    async _handleGetDocument(request, uid) {
        const work = (instance) => {
            return {
                doc: instance.collabState.doc.toJSON(),
                users: instance.userCount,
                version: instance.collabState.version,
                managerId: this.managerId,
            };
        };
        return this._toResponse(Either.map(await this._getInstance(request.body.docName), work), request.type);
    }
    async _handlePullEvents(request, uid) {
        const work = (instance) => {
            if (this.managerId !== request.body.managerId) {
                return Either.left(CollabFail.IncorrectManager);
            }
            return instance.getEvents(request.body);
        };
        return this._toResponse(Either.flatMap(await this._getInstance(request.body.docName), work), request.type);
    }
    async _handlePushEvents(request, uid) {
        const work = (instance) => {
            const { type, body } = request;
            log(`uid=${uid} userId=${body.userId} ${type} payload-steps=${JSON.stringify(body.steps)}`);
            if (this.managerId !== body.managerId) {
                return Either.left(CollabFail.IncorrectManager);
            }
            const result = instance.addEvents(body);
            if (Either.isRight(result)) {
                queueMicrotask(() => {
                    var _a;
                    (_a = this._serverCom) === null || _a === void 0 ? void 0 : _a.broadcast({
                        type: CollabManagerBroadCastType.NewVersion,
                        body: {
                            docName: instance.docName,
                            version: instance.collabState.version,
                        },
                    });
                });
            }
            return result;
        };
        return this._toResponse(Either.flatMap(await this._getInstance(request.body.docName), work), request.type);
    }
    _toResponse(val, requestType) {
        if (Either.isLeft(val)) {
            return {
                ok: false,
                body: val.left,
                type: requestType,
            };
        }
        else {
            return {
                ok: true,
                body: val.right,
                type: requestType,
            };
        }
    }
}
class Instance {
    constructor(docName, schema, _collabState, _applyState = (docName, newCollabState, oldCollabState) => {
        return true;
    }) {
        this.docName = docName;
        this.schema = schema;
        this._collabState = _collabState;
        this._applyState = _applyState;
        this.lastActive = Date.now();
    }
    get collabState() {
        return this._collabState;
    }
    // TODO add userCount
    get userCount() {
        return 1;
    }
    addEvents({ clientID, version: rawVersion, steps, userId, docName, }) {
        this.lastActive = Date.now();
        let version = nonNegInteger(rawVersion);
        if (version === undefined) {
            return Either.left(CollabFail.InvalidVersion);
        }
        const parsedSteps = steps.map((s) => Step.fromJSON(this.schema, s));
        return Either.flatMap(CollabServerState.addEvents(this._collabState, version, parsedSteps, clientID), (collabState) => {
            if (this._applyState(docName, collabState, this._collabState)) {
                this._collabState = collabState;
            }
            else {
                return Either.left(CollabFail.ApplyFailed);
            }
            return Either.right({
                empty: null,
            });
        });
    }
    getEvents({ docName, version, userId, managerId, }) {
        return Either.map(CollabServerState.getEvents(this._collabState, version), (events) => {
            return {
                version: events.version,
                steps: events.steps.map((step) => step.toJSON()),
                clientIDs: events.steps.map((step) => step.clientID),
                users: this.userCount, // TODO
            };
        });
    }
}
function nonNegInteger(str) {
    let num = Number(str);
    if (!isNaN(num) && Math.floor(num) === num && num >= 0) {
        return num;
    }
    return undefined;
}

export { CollabManager, CollabServerState, DEFAULT_MANAGER_ID, MAX_STEP_HISTORY };
